@@ -60,6 +60,7 @@ static GSList *adapters;
 static gboolean update_adapter(struct hdp_adapter *adapter);
 static struct hdp_device *create_health_device(DBusConnection *conn,
 						struct btd_device *device);
+static void free_echo_data(struct hdp_echo_data *edata);
 
 struct hdp_create_dc {
 	DBusConnection			*conn;
@@ -81,17 +82,56 @@ struct hdp_tmp_dc_data {
 };
 
 struct hdp_echo_data {
-	int			wid;		/* Watcher for echo channels */
 	gboolean		echo_done;	/* Is a echo was already done */
 	gpointer		buf;		/* echo packet sent */
 	uint			tid;		/* echo timeout */
-	struct hdp_tmp_dc_data	*hdp_conn;	/* temporal data */
 };
 
+static struct hdp_channel *hdp_channel_ref(struct hdp_channel *chan)
+{
+	if (!chan)
+		return NULL;
+
+	chan->ref++;
+
+	DBG("health_channel_ref(%p): ref=%d", chan, chan->ref);
+	return chan;
+}
+
+static void free_health_channel(struct hdp_channel *chan)
+{
+	if (chan->mdep == HDP_MDEP_ECHO) {
+		free_echo_data(chan->edata);
+		chan->edata = NULL;
+	}
+
+	mcap_mdl_unref(chan->mdl);
+	hdp_application_unref(chan->app);
+	health_device_unref(chan->dev);
+	g_free(chan->path);
+	g_free(chan);
+}
+
+static void hdp_channel_unref(struct hdp_channel *chan)
+{
+	if (!chan)
+		return;
+
+	chan->ref --;
+	DBG("health_channel_unref(%p): ref=%d", chan, chan->ref);
+
+	if (chan->ref > 0)
+		return;
+
+	free_health_channel(chan);
+}
+
 static void free_hdp_create_dc(struct hdp_create_dc *dc_data)
 {
 	dbus_message_unref(dc_data->msg);
 	dbus_connection_unref(dc_data->conn);
+	hdp_application_unref(dc_data->app);
+	health_device_unref(dc_data->dev);
 
 	g_free(dc_data);
 }
@@ -109,7 +149,7 @@ static void hdp_create_data_unref(struct hdp_create_dc *dc_data)
 {
 	dc_data->ref--;
 
-	DBG("hdp_create_data_ref(%p): ref=%d", dc_data, dc_data->ref);
+	DBG("hdp_create_data_unref(%p): ref=%d", dc_data, dc_data->ref);
 
 	if (dc_data->ref > 0)
 		return;
@@ -121,6 +161,7 @@ static void free_hdp_conn_dc(struct hdp_tmp_dc_data *data)
 {
 	dbus_message_unref(data->msg);
 	dbus_connection_unref(data->conn);
+	hdp_channel_unref(data->hdp_chann);
 
 	g_free(data);
 }
@@ -138,7 +179,7 @@ static void hdp_tmp_dc_data_unref(struct hdp_tmp_dc_data *data)
 {
 	data->ref--;
 
-	DBG("hdp_conn_data_ref(%p): ref=%d", data, data->ref);
+	DBG("hdp_conn_data_unref(%p): ref=%d", data, data->ref);
 
 	if (data->ref > 0)
 		return;
@@ -259,6 +300,7 @@ static void device_unref_mcl(struct hdp_device *hdp_device)
 	if (!hdp_device->mcl)
 		return;
 
+	mcap_close_mcl(hdp_device->mcl, FALSE);
 	mcap_mcl_unref(hdp_device->mcl);
 	hdp_device->mcl = NULL;
 	hdp_device->mcl_conn = FALSE;
@@ -281,21 +323,10 @@ static void free_health_device(struct hdp_device *device)
 	g_free(device);
 }
 
-static void free_application(struct hdp_application *app)
-{
-	if (app->dbus_watcher)
-		g_dbus_remove_watch(connection, app->dbus_watcher);
-
-	g_free(app->oname);
-	g_free(app->description);
-	g_free(app->path);
-	g_free(app);
-}
-
 static void remove_application(struct hdp_application *app)
 {
 	DBG("Application %s deleted", app->path);
-	free_application(app);
+	hdp_application_unref(app);
 
 	g_slist_foreach(adapters, (GFunc) update_adapter, NULL);
 }
@@ -333,20 +364,21 @@ static DBusMessage *manager_create_application(DBusConnection *conn,
 
 	name = dbus_message_get_sender(msg);
 	if (!name) {
-		free_application(app);
+		hdp_application_unref(app);
 		return g_dbus_create_error(msg,
 					ERROR_INTERFACE ".HealthError",
 					"Can't get sender name");
 	}
 
 	if (!set_app_path(app)) {
-		free_application(app);
+		hdp_application_unref(app);
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".HealthError",
 				"Can't get a valid id for the application");
 	}
 
 	app->oname = g_strdup(name);
+	app->conn = dbus_connection_ref(conn);
 
 	applications = g_slist_prepend(applications, app);
 
@@ -376,6 +408,12 @@ static DBusMessage *manager_destroy_application(DBusConnection *conn,
 
 	l = g_slist_find_custom(applications, path, cmp_app);
 
+	if (!l)
+		return g_dbus_create_error(msg,
+					ERROR_INTERFACE ".InvalidArguments",
+					"Invalid arguments in method call, "
+					"no such application");
+
 	app = l->data;
 	applications = g_slist_remove(applications, app);
 
@@ -386,7 +424,7 @@ static DBusMessage *manager_destroy_application(DBusConnection *conn,
 
 static void manager_path_unregister(gpointer data)
 {
-	g_slist_foreach(applications, (GFunc) free_application, NULL);
+	g_slist_foreach(applications, (GFunc) hdp_application_unref, NULL);
 
 	g_slist_free(applications);
 	applications = NULL;
@@ -423,6 +461,9 @@ static DBusMessage *channel_get_properties(DBusConnection *conn,
 	path = device_get_path(chan->dev->dev);
 	dict_append_entry(&dict, "Device", DBUS_TYPE_OBJECT_PATH, &path);
 
+	path = chan->app->path;
+	dict_append_entry(&dict, "Application", DBUS_TYPE_OBJECT_PATH, &path);
+
 	if (chan->config == HDP_RELIABLE_DC)
 		type = g_strdup("Reliable");
 	else
@@ -468,7 +509,7 @@ static void hdp_mdl_reconn_cb(struct mcap_mdl *mdl, GError *err, gpointer data)
 
 		/* Send abort request because remote side */
 		/* is now in PENDING state */
-		if (!mcap_mdl_abort(chan->mdl, abort_mdl_cb, chan, NULL,
+		if (!mcap_mdl_abort(chan->mdl, abort_mdl_cb, NULL, NULL,
 								&gerr)) {
 			error("%s", gerr->message);
 			g_error_free(gerr);
@@ -543,12 +584,12 @@ static void device_reconnect_mdl_cb(struct mcap_mdl *mdl, GError *err,
 					ERROR_INTERFACE ".HealthError",
 					"Cannot reconnect: %s", gerr->message);
 	g_dbus_send_message(dc_data->conn, reply);
+	hdp_tmp_dc_data_unref(dc_data);
 	g_error_free(gerr);
 	gerr = NULL;
 
 	/* Send abort request because remote side is now in PENDING state */
-	if (!mcap_mdl_abort(dc_data->hdp_chann->mdl, abort_mdl_cb,
-					dc_data->hdp_chann, NULL, &gerr)) {
+	if (!mcap_mdl_abort(mdl, abort_mdl_cb, NULL, NULL, &gerr)) {
 		error("%s", gerr->message);
 		g_error_free(gerr);
 	}
@@ -607,7 +648,7 @@ static DBusMessage *channel_acquire(DBusConnection *conn,
 	dc_data = g_new0(struct hdp_tmp_dc_data, 1);
 	dc_data->conn = dbus_connection_ref(conn);
 	dc_data->msg = dbus_message_ref(msg);
-	dc_data->hdp_chann = chan;
+	dc_data->hdp_chann = hdp_channel_ref(chan);
 
 	if (chan->dev->mcl_conn) {
 		reply = channel_acquire_continue(hdp_tmp_dc_data_ref(dc_data),
@@ -658,12 +699,6 @@ static void free_echo_data(struct hdp_echo_data *edata)
 	if (edata->tid)
 		g_source_remove(edata->tid);
 
-	if (edata->wid)
-		g_source_remove(edata->wid);
-
-	if (edata->hdp_conn)
-		hdp_tmp_dc_data_unref(edata->hdp_conn);
-
 	if (edata->buf)
 		g_free(edata->buf);
 
@@ -677,6 +712,9 @@ static void health_channel_destroy(void *data)
 	struct hdp_device *dev = hdp_chan->dev;
 
 	DBG("Destroy Health Channel %s", hdp_chan->path);
+	if (!g_slist_find(dev->channels, hdp_chan))
+		goto end;
+
 	dev->channels = g_slist_remove(dev->channels, hdp_chan);
 
 	if (hdp_chan->mdep != HDP_MDEP_ECHO)
@@ -688,6 +726,7 @@ static void health_channel_destroy(void *data)
 	if (hdp_chan == dev->fr) {
 		char *empty_path;
 
+		hdp_channel_unref(dev->fr);
 		dev->fr = NULL;
 		empty_path = "/";
 		emit_property_changed(dev->conn, device_get_path(dev->dev),
@@ -695,11 +734,8 @@ static void health_channel_destroy(void *data)
 					DBUS_TYPE_OBJECT_PATH, &empty_path);
 	}
 
-	if (hdp_chan->mdep == HDP_MDEP_ECHO)
-		free_echo_data(hdp_chan->edata);
-
-	g_free(hdp_chan->path);
-	g_free(hdp_chan);
+end:
+	hdp_channel_unref(hdp_chan);
 }
 
 static GDBusMethodTable health_channels_methods[] = {
@@ -719,26 +755,32 @@ static struct hdp_channel *create_channel(struct hdp_device *dev,
 {
 	struct hdp_channel *hdp_chann;
 
+	if (!dev)
+		return NULL;
+
 	hdp_chann = g_new0(struct hdp_channel, 1);
 	hdp_chann->config = config;
-	hdp_chann->dev = dev;
-	hdp_chann->mdl = mdl;
+	hdp_chann->dev = health_device_ref(dev);
 	hdp_chann->mdlid = mdlid;
-	hdp_chann->app = app;
 
-	if (app)
+	if (mdl)
+		hdp_chann->mdl = mcap_mdl_ref(mdl);
+
+	if (app) {
 		hdp_chann->mdep = app->id;
-	else
+		hdp_chann->app = hdp_application_ref(app);
+	} else
 		hdp_chann->edata = g_new0(struct hdp_echo_data, 1);
 
 	hdp_chann->path = g_strdup_printf("%s/chan%d",
 					device_get_path(hdp_chann->dev->dev),
 					hdp_chann->mdlid);
 
-	dev->channels = g_slist_append(dev->channels, hdp_chann);
+	dev->channels = g_slist_append(dev->channels,
+						hdp_channel_ref(hdp_chann));
 
 	if (hdp_chann->mdep == HDP_MDEP_ECHO)
-		return hdp_chann;
+		return hdp_channel_ref(hdp_chann);
 
 	if (!g_dbus_register_interface(dev->conn, hdp_chann->path,
 					HEALTH_CHANNEL,
@@ -750,7 +792,7 @@ static struct hdp_channel *create_channel(struct hdp_device *dev,
 		return NULL;
 	}
 
-	return hdp_chann;
+	return hdp_channel_ref(hdp_chann);
 }
 
 static void remove_channels(struct hdp_device *dev)
@@ -771,6 +813,9 @@ static void remove_channels(struct hdp_device *dev)
 
 static void close_device_con(struct hdp_device *dev, gboolean cache)
 {
+	if (!dev->mcl)
+		return;
+
 	mcap_close_mcl(dev->mcl, cache);
 	dev->mcl_conn = FALSE;
 
@@ -811,7 +856,7 @@ static gboolean serve_echo(GIOChannel *io_chan, GIOCondition cond,
 	int fd, len;
 
 	if (cond & (G_IO_ERR | G_IO_HUP | G_IO_NVAL)) {
-		chan->edata->wid = 0;
+		hdp_channel_unref(chan);
 		return FALSE;
 	}
 
@@ -828,7 +873,7 @@ static gboolean serve_echo(GIOChannel *io_chan, GIOCondition cond,
 
 fail:
 	close_device_con(chan->dev, FALSE);
-	chan->edata->wid = 0;
+	hdp_channel_unref(chan);
 	return FALSE;
 }
 
@@ -876,15 +921,15 @@ static void hdp_mcap_mdl_connected_cb(struct mcap_mdl *mdl, void *data)
 		return;
 
 	chan = dev->ndc;
-	chan->mdl = mdl;
-	dev->ndc = NULL;
+	chan->mdl = mcap_mdl_ref(mdl);
 
 	if (!g_slist_find(dev->channels, chan))
-		dev->channels = g_slist_prepend(dev->channels, chan);
+		dev->channels = g_slist_prepend(dev->channels,
+							hdp_channel_ref(chan));
 
 	if (!check_channel_conf(chan)) {
 		close_mdl(chan);
-		return;
+		goto end;
 	}
 
 	if (chan->mdep == HDP_MDEP_ECHO) {
@@ -893,15 +938,14 @@ static void hdp_mcap_mdl_connected_cb(struct mcap_mdl *mdl, void *data)
 
 		fd = mcap_mdl_get_fd(chan->mdl);
 		if (fd < 0)
-			return;
+			goto end;
 
 		chan->edata->echo_done = FALSE;
 		io = g_io_channel_unix_new(fd);
-		chan->edata->wid = g_io_add_watch(io,
-				G_IO_ERR | G_IO_HUP | G_IO_NVAL | G_IO_IN,
-				serve_echo, chan);
+		g_io_add_watch(io, G_IO_ERR | G_IO_HUP | G_IO_NVAL | G_IO_IN,
+				serve_echo, hdp_channel_ref(chan));
 		g_io_channel_unref(io);
-		return;
+		goto end;
 	}
 
 	g_dbus_emit_signal(dev->conn, device_get_path(dev->dev), HEALTH_DEVICE,
@@ -910,13 +954,17 @@ static void hdp_mcap_mdl_connected_cb(struct mcap_mdl *mdl, void *data)
 					DBUS_TYPE_INVALID);
 
 	if (dev->fr)
-		return;
+		goto end;
 
-	dev->fr = chan;
+	dev->fr = hdp_channel_ref(chan);
 
 	emit_property_changed(dev->conn, device_get_path(dev->dev),
 					HEALTH_DEVICE, "MainChannel",
 					DBUS_TYPE_OBJECT_PATH, &dev->fr->path);
+
+end:
+	hdp_channel_unref(dev->ndc);
+	dev->ndc = NULL;
 }
 
 static void hdp_mcap_mdl_closed_cb(struct mcap_mdl *mdl, void *data)
@@ -956,10 +1004,11 @@ static void hdp_mcap_mdl_aborted_cb(struct mcap_mdl *mdl, void *data)
 	if (!dev->ndc)
 		return;
 
-	dev->ndc->mdl = mdl;
+	dev->ndc->mdl = mcap_mdl_ref(mdl);
 
 	if (!g_slist_find(dev->channels, dev->ndc))
-		dev->channels = g_slist_prepend(dev->channels, dev->ndc);
+		dev->channels = g_slist_prepend(dev->channels,
+						hdp_channel_ref(dev->ndc));
 
 	if (dev->ndc->mdep != HDP_MDEP_ECHO)
 		g_dbus_emit_signal(dev->conn, device_get_path(dev->dev),
@@ -967,6 +1016,7 @@ static void hdp_mcap_mdl_aborted_cb(struct mcap_mdl *mdl, void *data)
 					DBUS_TYPE_OBJECT_PATH, &dev->ndc->path,
 					DBUS_TYPE_INVALID);
 
+	hdp_channel_unref(dev->ndc);
 	dev->ndc = NULL;
 }
 
@@ -986,8 +1036,7 @@ static uint8_t hdp_mcap_mdl_conn_req_cb(struct mcap_mcl *mcl, uint8_t mdepid,
 
 	DBG("Data channel request");
 
-	if (mdepid == 0) {
-		/* Is an echo channel */
+	if (mdepid == HDP_MDEP_ECHO) {
 		switch (*conf) {
 		case HDP_NO_PREFERENCE_DC:
 			*conf = HDP_RELIABLE_DC;
@@ -1011,6 +1060,9 @@ static uint8_t hdp_mcap_mdl_conn_req_cb(struct mcap_mcl *mcl, uint8_t mdepid,
 		}
 
 		dev->ndc = create_channel(dev, *conf, NULL, mdlid, NULL, NULL);
+		if (!dev->ndc)
+			return MCAP_MDL_BUSY;
+
 		return MCAP_SUCCESS;
 	}
 
@@ -1094,7 +1146,7 @@ static uint8_t hdp_mcap_mdl_reconn_req_cb(struct mcap_mdl *mdl, void *data)
 		return MCAP_MDL_BUSY;
 	}
 
-	dev->ndc = chan;
+	dev->ndc = hdp_channel_ref(chan);
 
 	return MCAP_SUCCESS;
 }
@@ -1103,6 +1155,9 @@ gboolean hdp_set_mcl_cb(struct hdp_device *device, GError **err)
 {
 	gboolean ret;
 
+	if (!device->mcl)
+		return FALSE;
+
 	ret = mcap_mcl_set_cb(device->mcl, device, err,
 		MCAP_MDL_CB_CONNECTED, hdp_mcap_mdl_connected_cb,
 		MCAP_MDL_CB_CLOSED, hdp_mcap_mdl_closed_cb,
@@ -1240,9 +1295,11 @@ static void release_adapter_instance(struct hdp_adapter *hdp_adapter)
 {
 	if (!hdp_adapter->mi)
 		return;
-	mcap_release_instance(hdp_adapter->mi);
-	hdp_adapter->mi = NULL;
+
 	check_devices_mcl();
+	mcap_release_instance(hdp_adapter->mi);
+	mcap_instance_unref(hdp_adapter->mi);
+	hdp_adapter->mi = NULL;
 }
 
 static gboolean update_adapter(struct hdp_adapter *hdp_adapter)
@@ -1354,10 +1411,12 @@ static void delete_echo_channel(struct hdp_channel *chan)
 		return;
 	}
 
-	if (mcap_delete_mdl(chan->mdl, delete_echo_channel_cb, chan, NULL,
-									&err))
+	if (mcap_delete_mdl(chan->mdl, delete_echo_channel_cb,
+				hdp_channel_ref(chan),
+				(GDestroyNotify) hdp_channel_unref, &err))
 		return;
 
+	hdp_channel_unref(chan);
 	error("Error deleting the echo channel: %s", err->message);
 	g_error_free(err);
 
@@ -1374,7 +1433,8 @@ static void abort_echo_channel_cb(GError *err, gpointer data)
 			/* MDL is removed from MCAP so we can */
 			/* free the data channel without sending */
 			/* a MD_DELETE_MDL_REQ */
-			health_channel_destroy(chan);
+			/* TODO review the above comment */
+			/* hdp_channel_unref(chan); */
 		}
 		return;
 	}
@@ -1435,9 +1495,6 @@ end:
 	g_dbus_send_message(hdp_conn->conn, reply);
 	g_source_remove(edata->tid);
 	edata->tid = 0;
-	edata->wid = 0;
-	hdp_tmp_dc_data_unref(edata->hdp_conn);
-	edata->hdp_conn = NULL;
 	g_free(edata->buf);
 	edata->buf = NULL;
 
@@ -1445,6 +1502,7 @@ end:
 		close_device_con(chan->dev, FALSE);
 	else
 		delete_echo_channel(chan);
+	hdp_tmp_dc_data_unref(hdp_conn);
 
 	return FALSE;
 }
@@ -1456,12 +1514,15 @@ static gboolean echo_timeout(gpointer data)
 	int fd;
 
 	error("Error: Echo request timeout");
+	chan->edata->tid = 0;
+
 	fd = mcap_mdl_get_fd(chan->mdl);
 	if (fd < 0)
 		return FALSE;
 
 	io = g_io_channel_unix_new(fd);
 	g_io_channel_shutdown(io, TRUE, NULL);
+
 	return FALSE;
 }
 
@@ -1485,9 +1546,12 @@ static void hdp_echo_connect_cb(struct mcap_mdl *mdl, GError *err,
 		/* side is now in PENDING state. */
 		if (!mcap_mdl_abort(hdp_conn->hdp_chann->mdl,
 					abort_echo_channel_cb,
-					hdp_conn->hdp_chann, NULL, &gerr)) {
+					hdp_channel_ref(hdp_conn->hdp_chann),
+					(GDestroyNotify) hdp_channel_unref,
+					&gerr)) {
 			error("%s", gerr->message);
 			g_error_free(gerr);
+			hdp_channel_unref(hdp_conn->hdp_chann);
 		}
 		return;
 	}
@@ -1507,13 +1571,13 @@ static void hdp_echo_connect_cb(struct mcap_mdl *mdl, GError *err,
 	send_echo_data(fd, edata->buf, HDP_ECHO_LEN);
 
 	io = g_io_channel_unix_new(fd);
-	edata->hdp_conn = hdp_tmp_dc_data_ref(hdp_conn);
-	edata->wid = g_io_add_watch(io,
-			G_IO_ERR | G_IO_HUP | G_IO_NVAL | G_IO_IN,
-			check_echo, hdp_conn);
+	g_io_add_watch(io, G_IO_ERR | G_IO_HUP | G_IO_NVAL | G_IO_IN,
+			check_echo, hdp_tmp_dc_data_ref(hdp_conn));
 
-	edata->tid = g_timeout_add_seconds(ECHO_TIMEOUT, echo_timeout,
-							hdp_conn->hdp_chann);
+	edata->tid  = g_timeout_add_seconds_full(G_PRIORITY_DEFAULT,
+					ECHO_TIMEOUT, echo_timeout,
+					hdp_channel_ref(hdp_conn->hdp_chann),
+					(GDestroyNotify) hdp_channel_unref);
 
 	g_io_channel_unref(io);
 }
@@ -1538,7 +1602,7 @@ static void abort_and_del_mdl_cb(GError *err, gpointer data)
 		}
 	}
 
-	if (!mcap_delete_mdl(mdl, delete_mdl_cb, mdl, NULL, &gerr)) {
+	if (!mcap_delete_mdl(mdl, delete_mdl_cb, NULL, NULL, &gerr)) {
 		error("%s", gerr->message);
 		g_error_free(gerr);
 	}
@@ -1561,7 +1625,7 @@ static void hdp_mdl_conn_cb(struct mcap_mdl *mdl, GError *err, gpointer data)
 
 		/* Send abort request because remote side */
 		/* is now in PENDING state */
-		if (!mcap_mdl_abort(hdp_chann->mdl, abort_mdl_cb, hdp_chann,
+		if (!mcap_mdl_abort(hdp_chann->mdl, abort_mdl_cb, NULL,
 								NULL, &gerr)) {
 			error("%s", gerr->message);
 			g_error_free(gerr);
@@ -1582,7 +1646,7 @@ static void hdp_mdl_conn_cb(struct mcap_mdl *mdl, GError *err, gpointer data)
 	if (dev->fr)
 		return;
 
-	dev->fr = hdp_chann;
+	dev->fr = hdp_channel_ref(hdp_chann);
 
 	emit_property_changed(dev->conn, device_get_path(dev->dev),
 					HEALTH_DEVICE, "MainChannel",
@@ -1659,8 +1723,7 @@ static void device_create_mdl_cb(struct mcap_mdl *mdl, uint8_t conf,
 	hdp_tmp_dc_data_unref(hdp_conn);
 
 	/* Send abort request because remote side is now in PENDING state */
-	if (!mcap_mdl_abort(hdp_chan->mdl, abort_mdl_cb, hdp_chan, NULL,
-								&gerr)) {
+	if (!mcap_mdl_abort(mdl, abort_mdl_cb, NULL, NULL, &gerr)) {
 		error("%s", gerr->message);
 		g_error_free(gerr);
 	}
@@ -1678,9 +1741,11 @@ fail:
 	/* Send abort request because remote side is now in PENDING */
 	/* state. Then we have to delete it because we couldn't */
 	/* register the HealthChannel interface */
-	if (!mcap_mdl_abort(mdl, abort_and_del_mdl_cb, mdl, NULL, &gerr)) {
+	if (!mcap_mdl_abort(mdl, abort_and_del_mdl_cb, mcap_mdl_ref(mdl), NULL,
+								&gerr)) {
 		error("%s", gerr->message);
 		g_error_free(gerr);
+		mcap_mdl_unref(mdl);
 	}
 }
 
@@ -1698,16 +1763,23 @@ static void device_create_dc_cb(gpointer user_data, GError *err)
 		return;
 	}
 
+	if (!data->dev->mcl) {
+		g_set_error(&gerr, HDP_ERROR, HDP_CONNECTION_ERROR,
+				"Mcl was closed");
+		goto fail;
+	}
+
 	hdp_create_data_ref(data);
 
 	if (mcap_create_mdl(data->dev->mcl, data->mdep, data->config,
 						device_create_mdl_cb, data,
 						destroy_create_dc_data, &gerr))
 		return;
+	hdp_create_data_unref(data);
 
+fail:
 	reply = g_dbus_create_error(data->msg, ERROR_INTERFACE ".HealthError",
 							"%s", gerr->message);
-	hdp_create_data_unref(data);
 	g_error_free(gerr);
 	g_dbus_send_message(data->conn, reply);
 }
@@ -1721,7 +1793,7 @@ static DBusMessage *device_echo(DBusConnection *conn,
 	GError *err = NULL;
 
 	data = g_new0(struct hdp_create_dc, 1);
-	data->dev = device;
+	data->dev = health_device_ref(device);
 	data->mdep = HDP_MDEP_ECHO;
 	data->config = HDP_RELIABLE_DC;
 	data->msg = dbus_message_ref(msg);
@@ -1729,7 +1801,7 @@ static DBusMessage *device_echo(DBusConnection *conn,
 	data->cb = hdp_echo_connect_cb;
 	hdp_create_data_ref(data);
 
-	if (device->mcl_conn) {
+	if (device->mcl_conn && device->mcl) {
 		if (mcap_create_mdl(device->mcl, data->mdep, data->config,
 						device_create_mdl_cb, data,
 						destroy_create_dc_data, &err))
@@ -1840,9 +1912,9 @@ static DBusMessage *device_create_channel(DBusConnection *conn,
 					"channel should be reliable");
 
 	data = g_new0(struct hdp_create_dc, 1);
-	data->dev = device;
+	data->dev = health_device_ref(device);
 	data->config = config;
-	data->app = app;
+	data->app = hdp_application_ref(app);
 	data->msg = dbus_message_ref(msg);
 	data->conn = dbus_connection_ref(conn);
 	data->cb = hdp_mdl_conn_cb;
@@ -1936,7 +2008,7 @@ static DBusMessage *device_destroy_channel(DBusConnection *conn,
 	del_data = g_new0(struct hdp_tmp_dc_data, 1);
 	del_data->msg = dbus_message_ref(msg);
 	del_data->conn = dbus_connection_ref(conn);
-	del_data->hdp_chann = hdp_chan;
+	del_data->hdp_chann = hdp_channel_ref(hdp_chan);
 
 	if (device->mcl_conn) {
 		if (mcap_delete_mdl(hdp_chan->mdl, hdp_mdl_delete_cb,
@@ -1997,9 +2069,13 @@ static void health_device_destroy(void *data)
 						device_get_path(device->dev));
 
 	remove_channels(device);
+	if (device->ndc) {
+		hdp_channel_unref(device->ndc);
+		device->ndc = NULL;
+	}
 
 	devices = g_slist_remove(devices, device);
-	free_health_device(device);
+	health_device_unref(device);
 }
 
 static GDBusMethodTable health_device_methods[] = {
@@ -2034,8 +2110,9 @@ static struct hdp_device *create_health_device(DBusConnection *conn,
 	dev = g_new0(struct hdp_device, 1);
 	dev->conn = dbus_connection_ref(conn);
 	dev->dev = btd_device_ref(device);
-	l = g_slist_find_custom(adapters, adapter, cmp_adapter);
+	health_device_ref(dev);
 
+	l = g_slist_find_custom(adapters, adapter, cmp_adapter);
 	if (!l)
 		goto fail;
 
@@ -2054,7 +2131,7 @@ static struct hdp_device *create_health_device(DBusConnection *conn,
 	return dev;
 
 fail:
-	free_health_device(dev);
+	health_device_unref(dev);
 	return NULL;
 }
 
@@ -2119,3 +2196,24 @@ void hdp_manager_stop()
 	dbus_connection_unref(connection);
 	DBG("Stopped Health manager");
 }
+
+struct hdp_device *health_device_ref(struct hdp_device *hdp_dev)
+{
+	hdp_dev->ref++;
+
+	DBG("health_device_ref(%p): ref=%d", hdp_dev, hdp_dev->ref);
+
+	return hdp_dev;
+}
+
+void health_device_unref(struct hdp_device *hdp_dev)
+{
+	hdp_dev->ref--;
+
+	DBG("health_device_unref(%p): ref=%d", hdp_dev, hdp_dev->ref);
+
+	if (hdp_dev->ref > 0)
+		return;
+
+	free_health_device(hdp_dev);
+}
